#include "CRC16.h"

#define BAUDRATE 57600
#define LEDPORT 13

#define REQUEST_FRAME_LENGTH 9
#define RESPONSE_FRAME_LENGTH 11

// request frame format
// #aa$b$cc%
// aa - command - 2 bytes
// b - data - 1 byte
// cc - crc - 2 bytes
// valid frame: #12$1$%
uint8_t requestFrame[REQUEST_FRAME_LENGTH] = { 0 };

enum requestCommand
{
	UNKNOWN,
	GETSTATUS,
	SETSTATE
};

struct request
{
	enum requestCommand command;
	uint8_t data;
};

// response frame format
// #aaaa$b$cc%
// aaaa - current board time in millis - 4 bytes
// b - current led state - 1 byte
// cc - crc - 2 bytes
uint8_t responseFrame[RESPONSE_FRAME_LENGTH] = { 0 };

void pushToArray(uint8_t* arr, uint8_t lastElement)
{
	for (uint8_t i = 0; i < REQUEST_FRAME_LENGTH - 1; i++)
		arr[i] = arr[i + 1];
	arr[REQUEST_FRAME_LENGTH - 1] = lastElement;
}

uint8_t state = 0;

bool checkCrc()
{
	uint8_t crcHi = REQUEST_FRAME_LENGTH - 3;
	uint8_t crcLo = REQUEST_FRAME_LENGTH - 2;

	uint8_t frameCopy[REQUEST_FRAME_LENGTH];
	memcpy(frameCopy, requestFrame, REQUEST_FRAME_LENGTH);
	uint16_t frameCrc = (frameCopy[crcHi] << 8) | frameCopy[crcLo];
	frameCopy[crcHi] = 0;
	frameCopy[crcLo] = 0;
	uint16_t desiredCrc = CRC16::Get(frameCopy, REQUEST_FRAME_LENGTH);
	return desiredCrc == frameCrc;
}

void setCrc()
{
	int startCrcIndex = 8;
}

bool isValidFrame()
{
	if (requestFrame[0] != '#')
		return false;

	if (requestFrame[REQUEST_FRAME_LENGTH - 1] != '%')
		return false;

	if (requestFrame[3] != '$' || requestFrame[REQUEST_FRAME_LENGTH - 4] != '$')
		return false;

	if (!checkCrc())
		return false;

	return true;
}

struct request parseFrame()
{
	enum requestCommand comm;
	uint8_t dat;

	// check command type
	char code[3] = { 0 };
	code[0] = requestFrame[1];
	code[1] = requestFrame[2];

	if (strcmp(code, "GS") == 0)
		comm = GETSTATUS;
	else if (strcmp(code, "SL") == 0)
		comm = SETSTATE;
	else
		comm = UNKNOWN;

	// extract data from request
	dat = requestFrame[4];

	struct request result;
	result.command = comm;
	result.data = dat;
	return result;
}

void prepareResponse()
{
	uint32_t currentTime = millis();

	responseFrame[0] = '#';
	// insert current time
	responseFrame[1] = currentTime >> 24;
	responseFrame[2] = currentTime >> 16;
	responseFrame[3] = currentTime >> 8;
	responseFrame[4] = currentTime;
	responseFrame[5] = '$';
	// current led state
	responseFrame[6] = state;
	responseFrame[7] = '$';
	responseFrame[8] = 0;
	responseFrame[9] = 0;
	responseFrame[10] = '%';
	// insert crc
	uint16_t crc = CRC16::Get(responseFrame, RESPONSE_FRAME_LENGTH);
	responseFrame[8] = crc >> 8;
	responseFrame[9] = crc;
}

void setState(uint8_t value)
{
	state = value;
	if (state)
		digitalWrite(LEDPORT, HIGH);
	else
		digitalWrite(LEDPORT, LOW);
}

void sendResponse()
{
	uint8_t i;
	for (i = 0; i < RESPONSE_FRAME_LENGTH; i++)
		Serial.print((char)responseFrame[i]);
}

void communication()
{
	while (Serial.available() > 0)
	{
		uint8_t incomingByte = Serial.read();
		pushToArray(requestFrame, incomingByte);
		if (isValidFrame())
		{
			struct request req = parseFrame();
			if (req.command == GETSTATUS || req.command == SETSTATE)
			{
				if (req.command == SETSTATE)
					setState(req.data);

				prepareResponse();
				sendResponse();
			}
		}
	}
}

void setup()
{
	Serial.begin(BAUDRATE);
	pinMode(13, OUTPUT);
	setState(0);
}

void loop()
{
	communication();
}
