﻿using Communication.Frames;
using System;
using System.Collections.Generic;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Communication
{
    public class SerialConnection
    {
        public delegate void ConnectionOpenedHandler(string portName);
        public delegate void UpdateFramesCountsHandler(int sentFrames, int receivedFrames, int crcErrors);
        public delegate void ResponseFrameReceivedHandler(ResponseFrame receivedFrame);

        public event ConnectionOpenedHandler ConnectionOpened;
        public event Action ConnectionClosed;
        public event ResponseFrameReceivedHandler ResponseFrameReceived;
        public event UpdateFramesCountsHandler UpdateFramesCounts;

        private object serialLock = new object();

        private SerialPort serial;
        private ResponseParser responseParser;
        private RequestFormatter requestFormatter;

        private byte[] buffer;

        public int SentFrames { get; private set; } = 0;

        public int ReceivedFrames
        {
            get { return responseParser.ReceivedFrames; }
        }

        public int CrcErrors
        {
            get { return responseParser.CrcErrors; }
        }

        public bool IsOpened
        {
            get { return serial.IsOpen; }
        }

        public SerialConnection()
        {
            buffer = new byte[ResponseFrame.Length];
            responseParser = new ResponseParser();
            requestFormatter = new RequestFormatter();

            serial = new SerialPort
            {
                BaudRate = 57600,
                DtrEnable = true,
                RtsEnable = true,
                DataBits = 8,
                StopBits = StopBits.One,
                Handshake = Handshake.None,
                Parity = Parity.None
            };
        }

        private void Serial_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            int toRead = serial.BytesToRead;
            for (int i = 0;i < toRead; i++)
            {
                byte incomingByte = (byte)serial.ReadByte();
                buffer.Push(incomingByte);
                var frame = responseParser.Parse(buffer);
                if (frame != null)
                    ResponseFrameReceived?.Invoke(frame);
                UpdateFramesCounts?.Invoke(SentFrames, ReceivedFrames, CrcErrors);
            }
        }

        public void SendRequest(RequestFrame requestFrame)
        {
            lock(serialLock)
            {
                byte[] data = requestFormatter.GetData(requestFrame);
                serial.Write(data, 0, data.Length);
                SentFrames++;
                UpdateFramesCounts?.Invoke(SentFrames, ReceivedFrames, CrcErrors);
            }
        }

        public void Open(string portName)
        {
            serial.PortName = portName;
            serial.Open();
            serial.DataReceived += Serial_DataReceived;
            ConnectionOpened?.Invoke(portName);
        }

        public List<string> GetPorts()
        {
            var ports = SerialPort.GetPortNames().ToList();
            ports.Sort();
            return ports;
        }

        public void Close()
        {
            if (serial.IsOpen)
            {
                serial.DataReceived -= Serial_DataReceived;
                serial.Close();
                ConnectionClosed?.Invoke();
            }  
        }
    }
}
