﻿using Communication;
using Communication.Frames;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace PC
{
    public partial class MainForm : Form
    {
        private SerialConnection serialConnection;
        private DevicePolling devicePolling;
        private RepeatingTask repeatingTask;

        private System.Windows.Forms.Timer dummyTimer;
        private bool dummyState = false;

        public MainForm()
        {
            InitializeComponent();
            serialConnection = new SerialConnection();
            serialConnection.ResponseFrameReceived += SerialConnection_ResponseFrameReceived;
            serialConnection.ConnectionOpened += SerialConnection_ConnectionOpened;
            serialConnection.ConnectionClosed += SerialConnection_ConnectionClosed;
            serialConnection.UpdateFramesCounts += SerialConnection_UpdateFramesCounts;

            devicePolling = new DevicePolling(serialConnection);
            repeatingTask = new RepeatingTask();

            dummyTimer = new System.Windows.Forms.Timer();
            dummyTimer.Interval = 500;
            dummyTimer.Tick += DummyTimer_Tick;
        }

        private void DummyTimer_Tick(object sender, EventArgs e)
        {
            devicePolling.SetLedState(dummyState);
            dummyState = !dummyState;
        }

        private void SerialConnection_UpdateFramesCounts(int sentFrames, int receivedFrames, int crcErrors)
        {
            lbFramesStatus.Text = string.Format("Sent frames: {0}, received: {1}, CRC errors: {2}", sentFrames, receivedFrames, crcErrors);
        }

        private void SerialConnection_ConnectionClosed()
        {
            this.UpdateUI(() =>
            {
                lbStatus.Text = "Port closed";
                SetOpenButtonState(false);
            });
            repeatingTask.Stop();
        }

        private void SerialConnection_ConnectionOpened(string portName)
        {
            lbStatus.Text = "Port opened: " + portName;
            SetOpenButtonState(true);
            repeatingTask.Start(() =>
            {
                try
                {
                    devicePolling.Poll();
                }
                catch(Exception ex)
                {
                    AddMessage("Can not open port: " + ex.Message);
                    repeatingTask.Stop();
                }
            }, TimeSpan.FromMilliseconds(100));
        }

        private void SerialConnection_ResponseFrameReceived(ResponseFrame frame)
        {
            this.UpdateUI(() =>
            {
                tbState.Text = frame.LedState ? "ON" : "OFF";
                var time = TimeSpan.FromMilliseconds(frame.CurrentTime);
                tbTime.Text = string.Format("{0:00}:{1:00}:{2:00}.{3:000}", time.Hours, time.Minutes, time.Seconds, time.Milliseconds);
            });
        }

        private void MainForm_Load(object sender, EventArgs e)
        {
            RefreshSerialPorts();
        }

        private void RefreshSerialPorts()
        {
            var ports = serialConnection.GetPorts();
            cbPorts.DataSource = ports;
        }

        private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
        {
            repeatingTask.Stop();
            e.Cancel = true;
            // try to close form in other thread to avoid deadlock between serial's thread and UI
            var closeThread = new Thread(new ThreadStart(CloseInOtherThread));
            closeThread.Start();
        }

        private void CloseInOtherThread()
        {
            FormClosing -= MainForm_FormClosing;
            serialConnection.Close();
            this.UpdateUI(() => Close());
        }

        private void btnPortOpen_Click(object sender, EventArgs e)
        {
            try
            {
                if (serialConnection.IsOpened)
                    serialConnection.Close();
                else
                {
                    var selectedPort = (string)cbPorts.SelectedValue;
                    serialConnection.Open(selectedPort);
                }
            }
            catch(Exception ex)
            {
                AddMessage("Can not open port: " + ex.Message);
            }
        }

        private void btnOn_Click(object sender, EventArgs e)
        {
            AddMessage("Command sent: ON");
            try
            {
                devicePolling.SetLedState(true);
            }
            catch(Exception ex)
            {
                AddMessage("Can not send command: " + ex.Message);
            }
        }

        private void btnOff_Click(object sender, EventArgs e)
        {
            AddMessage("Command sent: OFF");
            try
            {
                devicePolling.SetLedState(false);
            }
            catch (Exception ex)
            {
                AddMessage("Can not send command: " + ex.Message);
            }
        }

        private void SetOpenButtonState(bool portOpened)
        {
            if (portOpened)
                btnPortOpen.Text = "Close";
            else
                btnPortOpen.Text = "Open";
        }

        private void AddMessage(string message)
        {
            this.UpdateUI(() =>
            {
                tbMessages.Text += message + Environment.NewLine;
                tbMessages.SelectionStart = tbMessages.TextLength;
                tbMessages.ScrollToCaret();
            });
        }

        private void MainForm_KeyDown(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Q)
                btnOn_Click(null, null);
            else if (e.KeyCode == Keys.W)
                btnOff_Click(null, null);
        }

        private void btnDummyTimer_Click(object sender, EventArgs e)
        {
            dummyTimer.Enabled = true;
        }
    }
}
